const { app, BrowserWindow, shell, Menu, ipcMain } = require(`electron`);
const path = require(`path`);
const fs = require(`fs`);
const request = require('request');
const DecompressZip = require(`decompress-zip`);

let mainWindow;
let downloadWindow;

var version = `1`;

var userDataPath = ``;
var versionPath = ``;
var savePath = ``;
var gamePath = ``;
var gameItself = ``;
var tmpPath = ``;
var downloadItem;

var isWindows = false;
var gameStartTimer;
var retryDownloadTimer;
var gameName = ``;
var latest = ``;
var visibleTimeouts = 1000;

var installPath = ``;
switch (process.platform) {
	case 'win32':
		gameName = `Penguin 3D/Penguin 3D.exe`;
		isWindows = true;
	break;
	case 'darwin':
		gameName = `Penguin 3D.app`;
	break;
	case 'linux':
		gameName = `Penguin 3D/Penguin 3D.x86`;
	break;
}

app.on(`ready`, () => {
	mainWindow = new BrowserWindow({
		width: 780,
		height: 490,
		titleBarStyle: process.platform == "darwin" ? "hidden" : "visible",
		//useContentSize: true,
		webPreferences: {
		  nodeIntegration: true,
		  contextIsolation: false,
		  enableRemoteModule: true,
		},
	});
	mainWindow.on(`close`, () => {
		app.quit();
	});
	
	ipcMain.on("startgame", (event, arg) => {
		checkForUpdates();
	});
	
	// DEBUG
	//mainWindow.webContents.openDevTools();
	
	downloadWindow = new BrowserWindow({parent: mainWindow, modal: true, show: false});
    
	userDataPath = app.getPath('documents');
	userDataPath = path.join(userDataPath, `Penguin3D`);
	versionPath = path.join(userDataPath, `version.txt`);
	gamePath = path.join(userDataPath, `Game`);
	tmpPath = path.join(userDataPath, `Temp`);
	savePath = path.join(tmpPath, `Penguin3D.zip`);
	gameItself = path.join(gamePath, gameName);
	
	setMainMenu();
	
	mainWindow.loadURL(path.join(`file://`, __dirname, `index.html`));
});

function setMainMenu() {
	var template = [
	{
		label: `Penguin3D`,
		submenu: [{
			label: 'Quit',
			accelerator: 'Cmd+Q',
			click() {
				app.quit();
			}
		}]
	}];
	
	Menu.setApplicationMenu(Menu.buildFromTemplate(template));
	
	if(isWindows)
		Menu.setApplicationMenu(null)
	
	downloadWindow.webContents.session.on('will-download', (event, item, webContents) => {
		downloadItem = item;
		
		clearTimeout(retryDownloadTimer);
		
		status(`Clearing temporary file...`);
			
		try {
			fs.unlinkSync(savePath);
		}catch(e) {
			status(`No such temp file.`);
		}
		
		status(`Download starting...`);
		
		item.setSavePath(savePath);
		
		item.on('updated', (event, state) => {
			if (state === 'interrupted') {
				status('Download was interrupted. It will resume when the connection is back.');
				setTimeout(resumeDownload, 1500);
			} else if (state === 'progressing') {
				if (item.isPaused()) {
					status(`Download is paused. Resuming...`);
					setTimeout(resumeDownload, 1500);
				} else {
					var prog = Math.floor(item.getReceivedBytes() / item.getTotalBytes() * 100);
					status(`<b3>Downloading Penguin3D ${latest}</b3><div id='path'>${savePath}</div><br><b1>${prog}%</b1>`);
				}
			}
		});
		item.once('done', (event, state) => {
			if (state === 'completed') {
				unzipGame();
			} else {
				status(`Download failed: ${state}`);
				failsafe();
			}
		})
	})
}
function checkForUpdates() {
	status(`Checking for updates...`);
	
	gameStartTimer = setTimeout(startGame, 10000);
	
	request.get(`https://install.penguin3d.net/Penguin3D/latest.php?source=launcher&lv=3&platform=` + process.platform + `&version=` + version, function (error, response, body) {
		console.log(error, response, body);
		if(!response) {
			response = {};
		}
		if (error || response.statusCode != 200) {
			status(`There was a problem checking for updates.<br>` + error);
			failsafe();
			return;
		}
		clearTimeout(gameStartTimer);
		var data = String(body);
		
		if(data.indexOf(`ERROR=`) > -1) {
			var errorData = data.split(`ERROR=`)[1];
			status(errorData);
			setTimeout(failsafe, 5000);
			return;
		}
		if(data.indexOf(`LAUNCHERNOTICE=`) > -1) {
			var launcherNotice = data.split(`LAUNCHERNOTICE=`)[1];
			status(launcherNotice);
			return;
		}
		
		var dataObj = {};
		
		try {
			dataObj = JSON.parse(data);
		}catch(e) {
			status(`Couldn't find the latest version. Retrying...`);
			setTimeout(failsafe, 2000);
			return;
		}
		
		installPath = dataObj[`P3DURL`];
		latest = dataObj[`P3DVER`];
		
		onRequestDone();
	});
}
function onRequestDone() {
	gameStartTimer = setTimeout(startGame, 10000);
	status(`Latest version: ` + latest + `<br>Checking current version...`);
	setTimeout(checkCurrentVersion, visibleTimeouts);
}
function checkCurrentVersion() {
	fs.readFile(versionPath, 'utf-8', (err, data) => {
		clearTimeout(gameStartTimer);
		if(err) {
			status(`No version found. <br>Installing Penguin3D...`);
			setTimeout(installNewVersion, visibleTimeouts);
			return;
		}
		var ver = data;
		if(ver != latest) {
			status(`Current: ` + ver + ` <br>Updating Penguin3D...`);
			setTimeout(installNewVersion, visibleTimeouts);
			return;
		}else{
			startGame();
		}
	});
}
function installNewVersion() {
	clearTimeout(retryDownloadTimer);
	
	status(`Installing Penguin3D ` + latest + `...`);
	
	downloadWindow.loadURL(installPath);
	
	retryDownloadTimer = setTimeout(installNewVersion, 10000);
}
function unzipGame() {
	status(`Extracting game...`);
	
	var unzipper = new DecompressZip(savePath);
	unzipper.on('error', function (err) {
		status('Extraction error: ' + err);
		failsafe();
	});
	unzipper.on('extract', function (log) {
		unzipFinished();
	});
	unzipper.on('progress', function (fileIndex, fileCount) {
		var prog = Math.floor((fileIndex + 1) / fileCount * 100);
		status(`<b3>Extracting ${latest}</b3><div id='path'>${gamePath}</div><br><b1>${prog}%</b1>`);
	});
	unzipper.extract({path: gamePath});
}
function unzipFinished() {
	status(`Deleting temporary files...`);
	fs.unlink(savePath, function(err) {
		if(err) {
			status(`Error deleting temp file at ` + savePath + `<br>` + err);
			failsafe();
			return;
		}
		status(`Writing new version number...`);
		fs.writeFile(versionPath, latest, function(err) {
			if(err) {
				status(`Error updating version file at ` + versionPath + `<br>` + err);
				failsafe();
				return;
			}
			startGame(true);
		});
	});
}
function startGame(fromUnzip) {
	status(`Starting Penguin3D`);
	
	//shell.openItem(gameItself);
	fs.exists(gameItself, function(exists){
		if(!exists) {
			status(`Game does not exist at location '` + gameItself + `'<br><br> Opening game path at '` + gamePath + `'.`);
			shell.openExternal('file://' + gamePath);
			if(!fromUnzip) {
				installNewVersion();
			}
			return;
		}
		shell.openExternal(`file://` + gameItself);
		app.quit();
	});
}
function failsafe() {
	gameStartTimer = setTimeout(startGame, 2000);
}
function resumeDownload() {
	downloadItem.resume();
}
function clearTemp() {
	fs.readdir(tmpPath, (err, files) => {
	if (err) throw err;
	
	for (const file of files) {
		fs.unlink(path.join(directory, file), err => {
		  if (err) throw err;
		});
	}
	});
}

function status(str) {
	mainWindow.webContents.executeJavaScript(`status(\`` + str + `\`);`);
}